//Canny.bsv
import Vector::*;
import FIFO::*;
import GetPut::*;
import ClientServer::*;
import Connectable::*;
import Types::*;

typedef enum {
        Load,
        Output,
        FirstLoad,
        Suppress,
        Sobel,
        SecondLoad,
        Process
} CannyStage deriving(Bits, Eq, FShow);

typedef Server#(
    Vector#(64, APixel),
    Vector#(64, Bit#(8))
) Canny#(numeric type h, numeric type w, numeric type thigh);

typedef struct {  Vector#(66, APixel) line1;
                  Vector#(66, APixel) line2;
				  Vector#(66, APixel) line3;
} Reg2Can deriving (Bits, Eq);

module mkCannyFilter(Canny#(h,w,thigh) ifc); 

	Bit#(8) higherthreshold = fromInteger(valueof(thigh));	
	Integer h_im =  valueOf(h); // height of image for calculations
    Integer w_im =  valueOf(w); //width
	Integer num_col = w_im/64; //number of chunks received in line
    
    Reg#(Bit#(16)) row_count <- mkReg(0); //current row 
	Reg#(Bit#(16)) col_count <- mkReg(0);	//current column 
    
    //REGISTERS ARE INT#11 BECAUSE OF SOBEL KERNEL MAX *4 + SIGNED VALUE +/-1024
    Reg#(Vector#(TDiv#(w,64),Vector#(64, APixel))) r1 <- mkReg(replicate(newVector())); //top line
    Reg#(Vector#(TDiv#(w,64),Vector#(64, APixel))) r2 <- mkReg(replicate(newVector())); //middle line
    Reg#(Vector#(TDiv#(w,64),Vector#(64, APixel))) r3 <- mkReg(replicate(newVector())); //bottom line
		   
	//TRY USING CFFIFO IN FUTURE !!!!!!!!!!!!!!!!!!!!!!!
    FIFO#(Vector#(64, APixel)) infifo <- mkFIFO(); 
    FIFO#(Vector#(64, Bit#(8))) outfifo <- mkFIFO();
	
	FIFO#(Reg2Can)  reg2can <- mkFIFO();	// intermediate fifo
											// input to canny filter stage
    
	rule getInput1(col_count< fromInteger(num_col));
 
		// if line not completely received						
		if (row_count< fromInteger(h_im)) begin
			let inp = infifo.first();	//get input
			infifo.deq();					
			
			r3[col_count] <= inp;
			
			if (row_count>0 && col_count>0) begin
				Vector#(66, APixel) line1 ;
				Vector#(66, APixel) line2;
				Vector#(66, APixel) line3;
				for (Integer i =0; i<64; i=i+1) begin
					line1[i+1] = r1[col_count-1][i] ;
					line2[i+1] = r2[col_count-1][i];
					line3[i+1] = r3[col_count-1][i];
				end
				line1[65]=r1[col_count][0];
				line2[65]=r2[col_count][0];
				line3[65]=inp[0];
				
				if (col_count==1) begin	
					line1[0]=r1[0][0];
					line2[0]=r2[0][0];
					line3[0]=r3[0][0];
				end else  begin
					line1[0]=r1[col_count-2][63];
					line2[0]=r2[col_count-2][63];
					line3[0]=r3[col_count-2][63];
				end			
				
				
				Reg2Can r2c= Reg2Can {line1: line1, line2: line2, line3:line3};       
				reg2can.enq(r2c);
			end 
			
		end
		col_count <= col_count+1;
	endrule 
	
	rule getInput2(col_count== fromInteger(num_col));

		col_count <= 0;
		row_count <= row_count+1;
		
		if (row_count>0) begin
			Vector#(66, APixel) line1;
			Vector#(66, APixel) line2;
			Vector#(66, APixel) line3;
			for (Integer i =0; i<64; i=i+1) begin
				line1[i+1] = r1[col_count-1][i];
				line2[i+1] = r2[col_count-1][i];
				line3[i+1] = r3[col_count-1][i];
			end
			line1[0]=r1[col_count-2][63];
			line2[0]=r3[col_count-2][63];
			line3[0]=r2[col_count-2][63];
			line1[65] = r1[col_count-1][63];
			line2[65] = r2[col_count-1][63];
			line3[65] = r3[col_count-1][63];
			Reg2Can r2c = Reg2Can {line1: line1, line2: line2, line3:line3}; 
			reg2can.enq(r2c);
		end
		
		if (row_count==0) begin
			r2<=r3;
			r1<=r3;
		end else begin
			r1<=r2;	//shift the registers
			r2<=r3;	//shift the registers
		end		
    endrule
    
    //RULE THAT MAKES POSSIBLE USING THIS FILTER FOR VIDEOS AS WELL AS FOR IMAGES
    rule reset_frame (row_count == fromInteger(h_im + 1) );
        row_count <= 0; 
    endrule
    
    rule compute_canny;
		let r2c = reg2can.first();
		reg2can.deq();
		Bit#(2) angle;
		Bit#(8) a1; Bit#(8) a2; Bit#(8) a3;
		Bit#(8) b1; Bit#(8) b2; Bit#(8) b3;
		Bit#(8) c1; Bit#(8) c2; Bit#(8) c3;
		
		Vector#(64, Bit#(8)) canny_out;
		
		for (Bit#(16) i=0; i<64; i=i+1) begin
			a1=r2c.line1[i].mag; a2=r2c.line1[i+1].mag; a3=r2c.line1[i+2].mag;
			b1=r2c.line2[i].mag; b2=r2c.line2[i+1].mag; b3=r2c.line2[i+2].mag;
			c1=r2c.line3[i].mag; c2=r2c.line3[i+1].mag; c3=r2c.line3[i+2].mag;
			angle = r2c.line2[i+1].angle;
			
			Bit#(8) out = suppress(a1, a2, a3, b1, b2, b3, c1, c2, c3, higherthreshold, angle);
			canny_out[i] = out;			
		end
		
		outfifo.enq(canny_out);
    endrule

	interface request  = toPut (infifo);
    interface response = toGet (outfifo);

endmodule

function Bit#(8) suppress(    Bit#(8) a1, Bit#(8) a2, Bit#(8) a3,
							Bit#(8) b1, Bit#(8) b2, Bit#(8) b3,
							Bit#(8) c1, Bit#(8) c2, Bit#(8) c3, 
							Bit#(8) thigh, Bit#(2) angle);

	Pixel out = case (angle)
	0: return ((b2 >= b1) && (b2 >= b3) && (b2 >= thigh) ) ? 255 : 0;
					
	1: return ((b2 >= c1) && (b2 >= a3) && (b2 >= thigh) ) ? 255 : 0;
					
	2: return ((b2 >= a2) && (b2 >= c2) && (b2 >= thigh) ) ? 255 : 0;
					
	3: return ((b2 >= a1) && (b2 >= c3) && (b2 >= thigh) ) ? 255 : 0;

	endcase;

	return out;
endfunction
