// Sobel.bsv -Version 1. 
// Changes - > 
// *takes 64 byte inputs and can work concurrently - taking input and  filtering at the same time
// * registers are 8 bit values 
// This sobel filter sketch

//import what is necessary
import Vector::*;
import ClientServer::*;
import FIFO::*;
import GetPut ::*;
//import FPGAMemory::*;

typedef Server#(
    Vector#(64, Bit#(8)),
    SobelOut#(64)
) Sobel#(numeric type h, numeric type w);

typedef struct {  Vector#(w, Bit#(8)) g;
                  Vector#(w, Bit#(2)) angle;
} SobelOut#(numeric type w) deriving (Bits, Eq);

typedef struct {  Vector#(66, Bit#(8)) line1;
                  Vector#(66, Bit#(8)) line2;
				  Vector#(66, Bit#(8)) line3;
} Reg2Sob deriving (Bits, Eq);

module mkSobel(Sobel#(h,w) ifc);
    Integer h_im =  valueOf(h); // height of image for calculations
    Integer w_im =  valueOf(w); //width
	Integer num_col = w_im/64; //number of chunks received in line
    
    Reg#(Bit#(16)) row_count <- mkReg(0); //current row - max number of rows 4096
	Reg#(Bit#(16)) col_count <- mkReg(0);	//current column - max width of image 4096
    Reg#(Bool) need_data <- mkReg(True);
    
    //REGISTERS ARE INT#11 BECAUSE OF SOBEL KERNEL MAX *4 + SIGNED VALUE +/-1024
    Reg#(Vector#(TDiv#(w,64),Vector#(64, Bit#(8)))) r1 <- mkReg(replicate(newVector())); //top line
    Reg#(Vector#(TDiv#(w,64),Vector#(64, Bit#(8)))) r2 <- mkReg(replicate(newVector())); //middle line
    Reg#(Vector#(TDiv#(w,64),Vector#(64, Bit#(8)))) r3 <- mkReg(replicate(newVector())); //bottom line
           
		   
	//TRY USING CFFIFO IN FUTURE !!!!!!!!!!!!!!!!!!!!!!!
    FIFO#(Vector#(64, Bit#(8))) infifo <- mkFIFO(); 
    FIFO#(SobelOut#(64)) outfifo <- mkFIFO();
	
	FIFO#(Reg2Sob)  reg2sob <- mkFIFO();	// intermediate fifo
												// input to sobel filter stage
    
    rule getInput1(col_count< fromInteger(num_col));
 
			// if line not completely received						
			if (row_count< fromInteger(h_im)) begin
				let inp = infifo.first();	//get input
				infifo.deq();					
				
				r3[col_count] <= inp;
				
				if (row_count>0 && col_count>0) begin
					Vector#(66, Bit#(8)) line1 ;
					Vector#(66, Bit#(8)) line2;
					Vector#(66, Bit#(8)) line3;
					for (Integer i =0; i<64; i=i+1) begin
						line1[i+1] = r1[col_count-1][i] ;
						line2[i+1] = r2[col_count-1][i];
						line3[i+1] = r3[col_count-1][i];
					end
					line1[65]=r1[col_count][0];
					line2[65]=r2[col_count][0];
					line3[65]=inp[0];
					
					if (col_count==1) begin	
						line1[0]=r1[0][0];
						line2[0]=r2[0][0];
						line3[0]=r3[0][0];
					end else  begin
						line1[0]=r1[col_count-2][63];
						line2[0]=r2[col_count-2][63];
						line3[0]=r3[col_count-2][63];
					end			
					
					
					Reg2Sob r2s = Reg2Sob {line1: line1, line2: line2, line3:line3};       
					reg2sob.enq(r2s);
				end 
				
			end
			col_count <= col_count+1;
	endrule 
	
	rule getInput2(col_count== fromInteger(num_col));

		col_count <= 0;
		row_count <= row_count+1;
		
		if (row_count>0) begin
			Vector#(66, Bit#(8)) line1;
			Vector#(66, Bit#(8)) line2;
			Vector#(66, Bit#(8)) line3;
			for (Integer i =0; i<64; i=i+1) begin
				line1[i+1] = r1[col_count-1][i];
				line2[i+1] = r2[col_count-1][i];
				line3[i+1] = r3[col_count-1][i];
			end
			line1[0]=r1[col_count-2][63];
			line2[0]=r3[col_count-2][63];
			line3[0]=r2[col_count-2][63];
			line1[65] = r1[col_count-1][63];
			line2[65] = r2[col_count-1][63];
			line3[65] = r3[col_count-1][63];
			Reg2Sob r2s = Reg2Sob {line1: line1, line2: line2, line3:line3}; 
			reg2sob.enq(r2s);
		end
		
		if (row_count==0) begin
			r2<=r3;
			r1<=r3;
		end else begin
			r1<=r2;	//shift the registers
			r2<=r3;	//shift the registers
		end		
    endrule
    
    //RULE THAT MAKES POSSIBLE USING THIS FILTER FOR VIDEOS AS WELL AS FOR IMAGES
    rule reset_frame (row_count == fromInteger(h_im + 1) );
        row_count <= 0; 
    endrule
    
    rule compute_sobel;
		let r2s = reg2sob.first();
		reg2sob.deq();
		Int#(11) gx;
		Int#(11) gy;
		Int#(11) a1; Int#(11) a2; Int#(11) a3;
		Int#(11) b1; Int#(11) b2; Int#(11) b3;
		Int#(11) c1; Int#(11) c2; Int#(11) c3;
		Bit#(11) mag_11;
		
		Vector#(64, Bit#(8)) mag;
		Vector#(64, Bit#(2)) ang;
		for (Bit#(16) i=0; i<64; i=i+1) begin
			a1=zeroExtend(unpack(r2s.line1[i])); a2=zeroExtend(unpack(r2s.line1[i+1]));
			a3=zeroExtend(unpack(r2s.line1[i+2])); b1=zeroExtend(unpack(r2s.line2[i]));
			b2=zeroExtend(unpack(r2s.line2[i+1])); b3=zeroExtend(unpack(r2s.line2[i+2]));
			c1=zeroExtend(unpack(r2s.line3[i])); c2=zeroExtend(unpack(r2s.line3[i+1]));
			c3=zeroExtend(unpack(r2s.line3[i+2]));
			
			
			gx = (a3-a1) +((b3-b1)<<1)+(c3-c1);
			gy = (c1-a1) +((c2-a2)<<1)+(c3-a3);
			
			mag_11 = pack(abs(gx)+abs(gy));
			if (mag_11>255) begin
				mag_11=255;
			end
			mag[i] = truncate(mag_11);
			
			Bit#(2) angle  = getAngle(gx,gy);
			ang[i] = angle;
			
		end
		
		SobelOut#(64) out = SobelOut {g: mag, angle:ang};
		outfifo.enq(out);
		// To be implemented
    endrule

    interface Put request = toPut(infifo);
    interface Get response = toGet(outfifo);

endmodule

//Calculate gradient direction by splitting space into 8 45degree angle segments
function Bit#(2) getAngle(Int#(11) gx, Int#(11) gy);
    Bit#(2) out ;
	Int#(16) abs_gx2 = abs(signExtend(gx)<<1);
	Int#(16) abs_gy2 = abs(signExtend(gy)<<1);
	Int#(16) abs_gx5 = abs(signExtend(gx)*5);
	Int#(16) abs_gy5 = abs(signExtend(gy)*5);
	
	if (abs_gy5 <= abs_gx2) begin
		out = 0;
	end else if (abs_gy2 >= abs_gx5) begin
		out = 2;
	end else begin
		if ((gx>0 && gy>0) || (gx<0 && gy<0)) begin
			out = 1;
		end else begin
			out = 3;
		end
	end
    return out;
endfunction