//Let's try to keep things in packages
package Tb;
import Vector::*;
import FIFO::*;
import Sobel::*;
import ClientServer::*;
import GetPut::*;
import Sobel::*;
import Canny::*;
import Types::*;
typedef Bit#(8) Pixel;


(* synthesize *)
module mkSobelTB(Sobel#(272,640));
    Sobel#(272,640) sobel <- mkSobel();
    return sobel;
endmodule

(* synthesize *)
module mkCannyTB(Canny#(272,640, 64));
	Canny#(272,640, 64) canny <- mkCannyFilter();
	return canny;
endmodule

(* synthesize *)
module mkTb (Empty);
  
  	Sobel#(272,640) sobel <- mkSobelTB();
	Canny#(272,640, 64) canny <- mkCannyFilter();
	
	Reg#(File) m_in <- mkRegU;
	Reg#(File) m_out_g <- mkRegU;
	Reg#(File) m_out_angle <- mkRegU;
	Reg#(File) m_out <- mkRegU;
  
  	Reg#(Int#(32)) line_num <- mkReg(0);
	Reg#(Bit#(2)) stage <- mkReg(0);
   
  	Integer w = 640;        //DEFINE WIDTH OF IMAGE
  	Integer h = 272;        //DEFINE HEIGHT OF IMAGE 
  
  	//OPEN FILES FOR READ AND WRITE
	rule inputs (stage == 0);
		File in <- $fopen("in.bin", "r");
		m_in <= in;
		File out1 <- $fopen("out_g.txt", "w");
		m_out_g <= out1;	
    		File out2 <- $fopen("out_angle.txt", "w");
		m_out_angle <= out2;				
		File whole_out <- $fopen("output.txt", "w");
		m_out <= whole_out;

		stage <= stage + 1;   
	endrule
	
	rule sobel_rule (stage != 0);
		Vector#(64, Bit#(8)) line;
		for (Integer i = 0; i < 64; i = i+1) begin
			let val <-  $fgetc(m_in);
			line[i] = pack(truncate(val)); 
		end
		sobel.request.put(line);  
	endrule
 
	rule canny_rule (stage != 0);
		Vector#(64, APixel) cannyin = ?;
		let sobel_out <- sobel.response.get();
		APixel apixel;
		
		for (Integer i = 0; i < 64; i = i + 1) begin
			apixel.mag = sobel_out.g[i];
			apixel.angle = sobel_out.angle[i];
			cannyin[i] = apixel;
			
			$fwriteb(m_out_g, apixel.mag);
			Bit#(8) angle_ext = zeroExtend(apixel.angle);
        	$fwriteb(m_out_angle, angle_ext);
		end

		canny.request.put(cannyin); 
   	endrule 

	rule write (stage != 0);
  		let cannyout <- canny.response.get();
		for (Integer j =0; j < 64; j = j + 1) begin	
			$fwriteb(m_out, cannyout[j]);
		end


		if (line_num == fromInteger(50*w*h/64-1)) begin
			stage <= 3;  
		end else begin
			line_num <= line_num + 1;
		end
	endrule
 
//always close the files
	rule close (stage == 3);
		$fclose(m_in);
		//$fclose(m_out_g);
    	//$fclose(m_out_angle);
		$fclose(m_out);
		$finish();
	endrule

endmodule: mkTb

endpackage: Tb

